"""WaitForIt provides a deferred descriptor.
Usage:
class Foo:
  bar = WaitForIt(func, *args, **kwargs)
Then:
foo = Foo()
foo.bar = Foo.bar.__get__(foo, Foo) = func(*args, **kwargs)

This allows lazy evaluation of the given function with the given
arguments. By passing 'THIS' as an argument, 'THIS' is replaced with the
instance passed to __get__.

By passing only a 'str' object as the first argument, __get__ returns:
getattr(instance, arg: str).

The first argument must be a callable, a str object pointing to a key or
THIS, which is redundant.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import re

from worktoy.parse import maybe
from worktoy.static import THIS, OWNER, ATTR
from worktoy.waitaminute import MissingVariable, TypeException
from worktoy.waitaminute import SubclassException
from worktoy.waitaminute import ReadOnlyError, ProtectedError

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias, Never
else:
  try:
    from types import FunctionType as Callable
  except ImportError:
    def func() -> None:
      pass


    Callable = type(func)


class WaitForIt:
  """Creates a deferred function that is called when the __get__ is first
  called. """

  #  Python API
  __field_name__ = None
  __field_owner__ = None

  #  Private variables
  __current_instance__ = None
  __current_owner__ = None
  __call_me_maybe__ = None
  __pos_args__ = None
  __key_args__ = None

  #  Getters
  def _getFieldName(self, ) -> str:
    """Get the field name."""
    if self.__field_name__ is None:
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  def _getFieldOwner(self, ) -> type:
    """Get the field owner."""
    if self.__field_owner__ is None:
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  def _getCurrentInstance(self, ) -> Any:
    """Get the current instance."""
    return self.__current_instance__

  def _getCurrentOwner(self, ) -> type:
    """Get the current owner."""
    currentOwner = maybe(self.__current_owner__, self.__field_owner__)
    if currentOwner is None:
      raise MissingVariable('__current_owner__', type)
    return currentOwner

  def _getCallMeMaybe(self, ) -> Callable:
    """Get the call me maybe function."""
    if self.__call_me_maybe__ is None:
      raise MissingVariable('__call_me_maybe__', Callable)
    if callable(self.__call_me_maybe__):
      return self.__call_me_maybe__
    raise TypeException(
        '__call_me_maybe__',
        self.__call_me_maybe__,
        Callable
    )

  def _getPosArgs(self, ) -> list:
    """Get the positional arguments."""
    posArgs = maybe(self.__pos_args__, [])
    if not isinstance(posArgs, (list, tuple)):
      raise TypeException('__pos_args__', posArgs, list, tuple)
    currentInstance = self._getCurrentInstance()
    if currentInstance is None:
      return posArgs
    currentOwner = self._getCurrentOwner()
    out = []
    for arg in posArgs:
      if arg is THIS:
        out.append(currentInstance)
        continue
      if arg is OWNER:
        out.append(currentOwner)
        continue
      if arg is ATTR:
        out.append(self)
        continue
      out.append(arg)
    return out

  def _getKeyArgs(self, ) -> dict:
    """Get the keyword arguments."""
    keyArgs = maybe(self.__key_args__, {})
    if not isinstance(keyArgs, dict):
      raise TypeException('__key_args__', keyArgs, dict)
    currentInstance = self._getCurrentInstance()
    if currentInstance is None:
      return keyArgs
    out = {}
    for key, value in keyArgs.items():
      if value is THIS:
        out[key] = currentInstance
        continue
      out[key] = value
    return out

  #  Setters
  def _setCurrentInstance(self, instance: Any) -> None:
    """Set the current instance."""
    if instance is not self.__current_instance__:
      self.__current_instance__ = instance

  def _setCurrentOwner(self, owner: type) -> None:
    """Set the current owner."""
    fieldOwner = self._getFieldOwner()
    if not issubclass(owner, fieldOwner):
      raise SubclassException(owner, fieldOwner)
    self.__current_owner__ = owner

  #  Others
  def _generatePrivateName(self, ) -> str:
    """Generates a private name for the field based on the field name. """
    fieldName = self._getFieldName()
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', fieldName).lower()

  #  Python API
  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: Any, owner: type, **kwargs) -> Any:
    """Get the value of the descriptor."""
    if instance is None:
      return self
    pvtName = self._generatePrivateName()
    if hasattr(instance, pvtName):
      return getattr(instance, pvtName)
    if kwargs.get('_recursion', False):
      raise RecursionError
    self._setCurrentInstance(instance)
    self._setCurrentOwner(owner)
    callMeMaybe = self._getCallMeMaybe()
    posArgs = self._getPosArgs()
    keyArgs = self._getKeyArgs()
    val = callMeMaybe(*posArgs, **keyArgs)
    setattr(instance, pvtName, val)
    return self.__get__(instance, owner, _recursion=True)

  def __set__(self, instance: Any, value: Any) -> Never:
    """The setter is disabled"""
    raise ReadOnlyError(instance, self, value)

  def __delete__(self, instance: Any) -> Never:
    """The deleter is disabled"""
    owner = self._getCurrentOwner()
    val = self.__get__(instance, owner)
    raise ProtectedError(self, instance, val)

  #  Constructors
  def __init__(self, *args: Any, **kwargs: Any) -> None:
    """Constructor for the WaitForIt class."""
    if not args:
      raise MissingVariable('__call_me_maybe__', Callable)
    if args[0] in [THIS, OWNER, ATTR]:
      self.__call_me_maybe__ = getattr
      self.__pos_args__ = (*args,)
    elif isinstance(args[0], str):
      self.__call_me_maybe__ = getattr
      self.__pos_args__ = (THIS, *args,)
    elif callable(args[0]):
      self.__call_me_maybe__ = args[0]
      self.__pos_args__ = (*args[1:],)
    else:
      raise TypeException('__call_me_maybe__', args[0], Callable)
    if kwargs:
      self.__key_args__ = kwargs



"""The auto function creates an enumeration member. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..attr import Field
from ..text import typeMsg
from ..waitaminute import MissingVariable, VariableNotNone

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class NUM:
  """The NUM class provides a temporary dataclass for enumeration
  entries. """

  __iter_contents__ = None

  __member_name__ = None
  __member_value__ = None

  key = Field()
  val = Field()

  @key.GET
  def _getKey(self, ) -> str:
    """Get the key of the enumeration member."""
    if self.__member_name__ is None:
      raise MissingVariable('__member_name__', str)
    if isinstance(self.__member_name__, str):
      return self.__member_name__
    raise TypeError(typeMsg('__member_name__', self.__member_name__, str))

  @key.SET
  def _setKey(self, key: str) -> None:
    """Set the key of the enumeration member."""
    if self.__member_name__ is not None:
      raise VariableNotNone('__member_name__', )
    if not isinstance(key, str):
      raise TypeError(typeMsg('__member_name__', key, str))
    self.__member_name__ = key

  @val.GET
  def _getVal(self, ) -> Any:
    """Get the value of the enumeration member."""
    if self.__member_value__ is None:
      return self.__member_name__
    return self.__member_value__

  @val.SET
  def _setVal(self, value: Any) -> None:
    """Set the value of the enumeration member."""
    if self.__member_value__ is not None:
      raise VariableNotNone('__member_value__', )
    self.__member_value__ = value

  def __iter__(self) -> Self:
    """Iterate over the enumeration member."""
    self.__iter_contents__ = [self.key, self.val]
    return self

  def __next__(self) -> Any:
    """Get the next item in the enumeration member."""
    if self.__iter_contents__:
      return self.__iter_contents__.pop(0)
    raise StopIteration

  def __getitem__(self, identifier: Any) -> Any:
    """Get the item from the enumeration member."""
    if isinstance(identifier, str):
      if len(identifier) > 2:
        if identifier.lower()[:3] == 'key':
          return self.key
        if identifier.lower()[:3] == 'val':
          return self.val
      raise KeyError(identifier)
    if isinstance(identifier, int):
      if identifier < 2:
        if identifier % 2:
          return self.val
        return self.key
    valType = type(self.val)
    if isinstance(identifier, valType):
      if identifier == self.val:
        return self.val

  def asDict(self) -> dict:
    """Get the enumeration member as a dictionary."""
    return {self.key: self.val}

  @classmethod
  def fromValue(cls, *args) -> Self:
    """Create a new enumeration member from a value."""
    self = cls()
    if len(args) == 1:
      value = args[0]
    else:
      value = args
    self.val = value
    return self


def auto(innerValue: Any = None, ) -> NUM:
  """The auto function creates an enumeration member.

  Args:
    innerValue: The value to be assigned to the enumeration member.
  """
  return NUM.fromValue(innerValue)



"""AttriBox provides a descriptor with lazy instantiation of the
underlying object. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
import re

from ..static import THIS
from ..text import typeMsg
from ..waitaminute import VariableNotNone, MissingVariable, TypeException
from . import AbstractDescriptor

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, Callable, TypeAlias

  Types: TypeAlias = tuple[type, ...]


class AbstractBox(AbstractDescriptor):
  """AttriBox provides a descriptor with lazy instantiation of the
  underlying object. """

  __field_type__ = None
  __pos_args__ = None
  __key_args__ = None

  def _getFieldTypes(self) -> Types:
    """Get the field type."""
    if self.__field_type__ is None:
      raise MissingVariable('__field_type__', type)
    if isinstance(self.__field_type__, tuple):
      for type_ in self.__field_type__:
        if not isinstance(type_, type):
          e = typeMsg('__field_type__', self.__field_type__, tuple)
          raise TypeError(e)
      else:
        return self.__field_type__
    raise TypeException('__field_type__', self.__field_type__, tuple)

  def _setFieldType(self, *args) -> None:
    """Set the field type."""
    if self.__field_type__ is not None:
      raise VariableNotNone('__field_type__')
    fieldTypes = []
    for type_ in args:
      if not isinstance(type_, type):
        e = typeMsg('type_', type_, type)
        raise TypeError(e)
      if type_ is float:
        fieldTypes.append(float)
        fieldTypes.append(int)
        continue
      if type_ is complex:
        fieldTypes.append(complex)
        fieldTypes.append(float)
        fieldTypes.append(int)
        continue
      fieldTypes.append(type_)
    self.__field_type__ = (*fieldTypes,)

  def _getPosArgs(self, instance: object = None) -> tuple:
    """Get the positional arguments."""
    posArgs = getattr(self, '__pos_args__', None)
    if posArgs is None:
      raise MissingVariable('__pos_args__', tuple)
    if not isinstance(posArgs, tuple):
      e = typeMsg('__pos_args__', posArgs, tuple)
      raise TypeError(e)
    if instance is None:
      return posArgs
    out = []
    for arg in posArgs:
      if arg is THIS:
        out.append(instance)
      else:
        out.append(arg)
    return (*out,)

  def _setPosArgs(self, *args) -> None:
    """Set the positional arguments."""
    self.__pos_args__ = (*args,)

  def _getKeyArgs(self, instance: object = None) -> dict:
    """Get the keyword arguments."""
    keyArgs = getattr(self, '__key_args__', None)
    if keyArgs is None:
      raise MissingVariable('__key_args__', dict)
    if not isinstance(keyArgs, dict):
      e = typeMsg('__key_args__', keyArgs, dict)
      raise TypeError(e)
    if instance is None:
      return keyArgs
    out = {}
    for key, value in keyArgs.items():
      if value is THIS:
        out[key] = instance
      else:
        out[key] = value
    return out

  def _setKeyArgs(self, **kwargs) -> None:
    """Set the keyword arguments."""
    self.__key_args__ = {**kwargs, }

  def getPrivateName(self, ) -> str:
    """Getter-function for the private name of the field."""
    fieldName = self.getFieldName()
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', fieldName).lower()

  @abstractmethod
  def _instanceGet(self, instance: object, **kwargs) -> object:
    """Get the instance of the descriptor."""

  @abstractmethod
  def _instanceSet(self, instance: object, value: object, **kwargs) -> None:
    """Set the instance of the descriptor."""

  @abstractmethod
  def _instanceDelete(self, instance: object, **kwargs) -> None:
    """Delete the instance of the descriptor."""



"""FlexBox provides a descriptor with lazy instantiation of the underlying
object and with type-flexibility. This means that the __set__ method will
attempt to cast to the field type, if the value is not of the field type."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import DELETED
from ..text import typeMsg, monoSpace
from ..waitaminute import MissingVariable, DeletedAttributeException
from worktoy.attr import AbstractBox

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class FlexBox(AbstractBox):
  """FlexBox provides a descriptor with lazy instantiation of the underlying
  object and with type-flexibility. This means that the __set__ method will
  attempt to cast to the field type, if the value is not of the field
  type."""

  def __class_getitem__(cls, fieldType: type) -> Self:
    """Get the field type."""
    self = cls()
    self._setFieldType(fieldType)
    return self

  def __call__(self, *args, **kwargs) -> Self:
    """Call the descriptor with the given arguments."""
    self._setPosArgs(*args)
    self._setKeyArgs(**kwargs)
    return self

  def _createInitObject(self, instance: object, ) -> Any:
    """Create the object."""
    fieldTypes = self._getFieldTypes()
    posArgs = self._getPosArgs(instance)
    keyArgs = self._getKeyArgs(instance)
    exception = None
    for cls in fieldTypes:
      try:
        newObject = cls(*posArgs, **keyArgs)
      except TypeError as typeError:
        if exception is None:
          exception = typeError
      else:
        return newObject
    else:
      raise exception

  def _createSetObject(self, instance: object, *args, **kwargs) -> Any:
    """Create the object."""
    fieldTypes = self._getFieldTypes()
    exception = None
    for cls in fieldTypes:
      try:
        newObject = cls(*args, )
      except TypeError as typeError:
        if exception is None:
          exception = typeError
      else:
        return newObject
    else:
      if args:
        if isinstance(args[0], (tuple, list)):
          if kwargs.get('_recursion', False):
            raise RecursionError
          return self._createSetObject(instance, *args[0], _recursion=True)
      raise exception

  def _getExistingObject(self, instance: object) -> Any:
    """Get the existing object."""
    pvtName = self.getPrivateName()
    existingObject = getattr(instance, pvtName, None)
    if existingObject is DELETED:
      e = """Attempted to access attribute '%s' from object of type: '%s',
      which has been deleted!""" % (pvtName, type(instance),)
      raise AttributeError(monoSpace(e))
    if existingObject is None:
      fieldType = self._getFieldTypes()
      raise MissingVariable(pvtName, fieldType)
    return existingObject

  def _instanceGet(self, instance: Any, **kwargs) -> Any:
    """Get the instance."""
    pvtName = self.getPrivateName()
    fieldTypes = self._getFieldTypes()
    try:
      existingObject = self._getExistingObject(instance)
    except MissingVariable:
      if kwargs.get('_recursion', False):
        raise RecursionError
      newObject = self._createInitObject(instance, )
      setattr(instance, pvtName, newObject)
      return self._instanceGet(instance, _recursion=True, )
    else:
      return existingObject

  def _instanceSet(self, instance: Any, value, *args, **kwargs) -> None:
    """Set the instance."""
    pvtName = self.getPrivateName()
    fieldTypes = self._getFieldTypes()
    if isinstance(value, fieldTypes):
      return setattr(instance, pvtName, value)
    if kwargs.get('_recursion', False):
      raise RecursionError
    newObject = self._createSetObject(instance, value, )
    return self._instanceSet(instance, newObject, _recursion=True, )

  def _instanceDelete(self, instance: object, **kwargs) -> None:
    """Delete the instance."""
    pvtName = self.getPrivateName()
    if getattr(instance, pvtName, None) is None:
      e = """Attempted to delete attribute '%s' from object of type: '%s', 
      which owns no such attribute!""" % (pvtName, type(instance),)
      raise AttributeError(monoSpace(e))
    setattr(instance, pvtName, DELETED)



"""NameHook filters named used in the namespace system. If it encounters
names that are near-misses of commonly used names, it will raise a
QuestionableSyntax exception. Because these names refer to built-in
functions, the near-miss names will not generally cause any errors related
to names. Instead, the built-in function is called instead of the
near-miss named custom function. Tracing the unexpected behaviour back
to the near-miss name is by no means obvious because of the presence
of the built-in function.

Below is a list of common functions along with potential near-miss names
and some comments.

# 1. __init__: Near-miss name: __int__
Both __init__ and __int__ find common use. The hook will inspect the
typehint of the return value. '__init__' and '__int__' should return
'None' and 'int' respectively. '__int__' should expect only 'self' as an
argument, whereas '__init__' might expect other arguments.

# 2. __set_name__: Near-miss name: __setname__ When a class body has a
key, value pair with value being of a type implementing __set_name__,
then this method is called when the owning class returns from __new__ in
the metaclass, but before __init__ is called. Although there is no
built-in implementation of __set_name__, no error will  be raised by its
absence. The typehints of the __set_name__ method  are:

  def __set_name__(self, owner: type, name: str) -> None:
    ...

# 3. __getitem__: Near-miss name: __get_item__ Dictionary like classes
call this method when encountering the syntax:
  dictLike = DictLike()
  print(dictLike[key])  # equivalent to:
  DictLike.__getitem__(dictLike, key)

# 4. __setitem__: Near-miss name: __set_item__ Exactly like above:
  dictLike = DictLike()
  dictLike[key] = value  # equivalent to:
  DictLike.__setitem__(dictLike, key, value)

# 5. __del__: Near-miss name: __delete__ When implementing the descriptor
protocol, the get and set accessors are '__get__' and '__set__' respectively.
The delete-accessor is '__delete__', but mistakenly using '__del__' does
happen. Since __del__ is a built-in function relating to garbage
collecting, this near-miss will cause HIGHLY UNDEFINED BEHAVIOUR.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AbstractHook
from ...waitaminute import QuestionableSyntax

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any
  from .. import AbstractNamespace as ASpace


class _Name:
  """Name descriptor."""

  __pvt_name__ = None

  def __init__(self, pvtName: str) -> None:
    """Initialize the Name object."""
    self.__pvt_name__ = pvtName

  def __get__(self, instance: object, owner: type) -> Any:
    """Get the name of the function."""
    if instance is None:
      return self
    return getattr(instance, self.__pvt_name__, )


class _NearMiss:
  """Encapsulates a near-miss name and the real name."""

  __correct_name__ = None
  __missed_name__ = None

  correct = _Name('__correct_name__')
  missed = _Name('__missed_name__')

  def __init__(self, correctName: str, missedName: str) -> None:
    """Initialize the NearMiss object."""
    self.__correct_name__ = correctName
    self.__missed_name__ = missedName


class NameHook(AbstractHook):
  """NameHook filters named used in the namespace system. If it encounters
  names that are near-misses of commonly used names, it will raise a
  QuestionableSyntax exception. """

  @classmethod
  def _getNearMisses(cls) -> list[_NearMiss]:
    """Get the near-miss names."""
    return [
        _NearMiss('__set_name__', '__setname__'),
        _NearMiss('__getitem__', '__get_item__'),
        _NearMiss('__setitem__', '__set_item__'),
        _NearMiss('__delitem__', '__del_item__'),
    ]

  @classmethod
  def _validateName(cls, name: str) -> bool:
    """Compares the name to list of potential near-miss names. If the name
    is a near-miss, a QuestionableSyntax exception is raised. """
    nearMisses = cls._getNearMisses()
    for nearMiss in nearMisses:
      if name == nearMiss.missed:
        raise QuestionableSyntax(nearMiss.missed, nearMiss.correct)
    return False

  def setItemHook(
      self,
      space: ASpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """
    return self._validateName(key)



"""OverloadHook hooks into the namespace system and collects the overload
decorated methods replacing them with a dispatcher that calls the
correct method based on the arguments passed to it. It is used to
provide a simple way to overload methods in the namespace system."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...static import Dispatch
from . import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias
  from worktoy.mcls import BaseSpace as BSpace
  from .. import FunctionType as Func

  Types: TypeAlias = tuple[type, ...]
  TypesDict: TypeAlias = dict[Types, Func]
  KeyTypesDict: TypeAlias = dict[str, TypesDict]
  FuncDict: TypeAlias = dict[str, list[Func]]
  HashDict: TypeAlias = dict[int, Func]
  KeyHashFunc: TypeAlias = dict[str, HashDict]
  FuncList: TypeAlias = list[Func]
  KeyFunc: TypeAlias = dict[str, Func]
  KeyFuncList: TypeAlias = dict[str, FuncList]


class OverloadHook(AbstractHook):
  """OverloadHook hooks into the namespace system and collects the overload
  decorated methods replacing them with a dispatcher that calls the
  correct method based on the arguments passed to it. It is used to
  provide a simple way to overload methods in the namespace system.
  """

  def setItemHook(
      self,
      space: BSpace,
      key: str,
      value: object,
      oldValue: object
  ) -> bool:
    """Set the item hook for the namespace system. This method is called
    when an item is set in the namespace system. It collects the
    overload decorated methods and replaces them with a dispatcher that
    calls the correct method based on the arguments passed to it."""
    if getattr(value, '__is_overloaded__', None) is None:
      return False
    typeSigs = getattr(value, '__type_sigs__', None)
    for sig in typeSigs:
      space.addOverload(key, sig, value)
    return True

  def postCompileHook(self, space: BSpace, namespace: dict) -> dict:
    """Post compile hook for the namespace system. This method is called
    after the namespace system is compiled. It collects the overload
    decorated methods and replaces them with a dispatcher that calls the
    correct method based on the arguments passed to it."""
    overloadMap = space.getOverloadMap()
    dispatchNames = []
    for key, sigMap in overloadMap.items():
      namespace[key] = Dispatch(sigMap)
      dispatchNames.append(key)
    namespace['__dispatch_names__'] = dispatchNames
    return namespace



"""VariableNotNone should be raised when a variable is unexpectedly not
None."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self


class VariableNotNone(Exception):
  """VariableNotNone should be raised when a variable is unexpectedly not
  None."""

  varName = _Attribute(None)

  def __init__(self, variableName: str = None) -> None:
    """Initialize the VariableNotNone object."""
    if variableName is None:
      Exception.__init__(self, 'Expected variable to be None')
    else:
      self.varName = variableName
      infoSpec = """Expected variable '%s' to be None"""
      Exception.__init__(self, monoSpace(infoSpec % variableName))



"""DeletedAttributeException is raised to indicate that an attribute
was previously deleted with a call to __delete__ but is now attempted to
be accessed. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import joinWords, monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any

  from ..attr import AbstractBox


class DeletedAttributeException(AttributeError):
  """DeletedAttributeException is raised to indicate that an attribute
  was previously deleted with a call to __delete__ but is now attempted to
  be accessed. """

  desc = _Attribute()
  instance = _Attribute()

  def __init__(self, desc: AbstractBox, instance: Any) -> None:
    """Initialize the DeletedAttributeException object."""
    self.desc = desc
    self.instance = instance
    fieldName = getattr(desc, '__field_name__', None)
    owner = type(instance)
    ownerName = owner.__name__
    clsName = type(desc).__name__
    infoSpec = """Tried accessing deleted attribute at: '%s.%s: %s' from 
    instance: '%s', but the attribute was already deleted!"""
    info = infoSpec % (ownerName, fieldName, clsName, repr(instance))
    AttributeError.__init__(self, monoSpace(info))

  def _resolveOther(self, other: Any) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the DeletedAttributeException object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.desc != other.desc:
        return False
      if self.instance != other.instance:
        return False
      return True
    return False



"""AbstractException provides an abstract baseclass for custom exceptions
in the 'worktoy.waitaminute' module."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import logging
from abc import abstractmethod

from worktoy.parse import maybe
from worktoy.waitaminute import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self


class AbstractException(Exception):
  """AbstractException provides an abstract baseclass for custom exceptions
  in the 'worktoy.waitaminute' module."""

  __class_attributes__ = None

  @classmethod
  def _getAttributes(cls) -> dict[str, _Attribute]:
    """Get the attributes of the class."""
    return maybe(cls.__class_attributes__, dict())

  def _resolveOther(self, other: Any) -> Self:
    """Resolve the other object to the same type as self."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (list, tuple)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the AbstractException object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)



"""The 'worktoy.keenum' module provides a baseclass for enumerations. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._auto import auto, NUM
from ._num_hook import NumHook
from ._num_space import NumSpace
from ._meta_num import MetaNum
from ._keenum import KeeNum


__all__ = [
    'auto',
    'NUM',
    'NumHook',
    'NumSpace',
    'MetaNum',
    'KeeNum',
]


"""KeeNum provides a baseclass for enumerations."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import MetaNum
from ..waitaminute import ReadOnlyError

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never
  from worktoy.mcls import FunctionType as Func


def _root(func: Func) -> Func:
  """Decorator to mark a function as a root function."""
  setattr(func, '__is_root__', True)
  return func


class KeeNum(metaclass=MetaNum):
  """KeeNum provides a baseclass for enumerations."""

  __slots__ = ('name', 'value', 'index')

  @_root
  def __init__(self, name: str, value: Any, index: int) -> None:
    """Initialize the KeeNum instance."""
    self.name = name
    self.value = value
    self.index = index

  @_root
  def __bool__(self, ) -> bool:
    """Return True if the value is not 0."""
    return False if self is type(self).NULL else True

  @_root
  def __str__(self, ) -> str:
    """Return the string representation of the KeeNum instance."""
    clsName = type(self).__name__
    return """%s.%s""" % (clsName, self.name.upper(),)

  @_root
  def __repr__(self, ) -> str:
    """Return the string representation of the KeeNum instance."""
    return """%s(%s, %s)""" % (type(self).__name__, self.name, self.value,)

  @_root
  def __eq__(self, other: object) -> bool:
    """Return True if the value is equal to the other value."""
    if not isinstance(other, KeeNum):
      return True if self.value == other else False
    return True if self is other else False

  @_root
  def __get__(self, instance: object, owner: type) -> KeeNum:
    """Return the KeeNum instance."""
    return self

  @_root
  def __set__(self, instance: object, value: object) -> Never:
    """Set the KeeNum instance."""
    raise ReadOnlyError(instance, self, value)

  @_root
  def __delete__(self, instance: object) -> Never:
    """Delete the KeeNum instance."""
    raise ReadOnlyError(instance, self, None)



"""ComplexNumber provides a class representation of the complex number
using features found in the 'worktoy' library. Briefly, a complex number
has a real and imaginary part both of which represented here as instances
of 'AttriBox'. The class inherits from the 'BaseObject' class allowing it
to overload functions based on types."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import sys
from math import atan2, cos, sin  # Provides necessary math functions

from worktoy.mcls import BaseObject
from worktoy.attr import AttriBox, Field
from worktoy.static import overload, THIS

try:
  from typing import Self
except ImportError:  # Older versions of Python
  Self = object  # Never used at runtime, just for type-hinting


class ComplexNumber(BaseObject):
  """
  Function overloading requires customization of the 'type' object
  itself. 'worktoy' provides 'BaseObject' which is derived from the
  'BaseMeta' class which implements the necessary methods for
  overloading.
  """

  REAL = AttriBox[float](0.0)  # Yes, really. AttriBox[TYPE](DEFAULT)
  IMAG = AttriBox[float](0.0)  # But linters may indicate warnings

  ABS = Field()  # Field is quite similar to property
  ARG = Field()  # Requiring accessor methods to be explicitly decorated.

  #  Constructor methods

  @overload(float, float)  # Instantiating from two floats
  @overload(float, int)
  @overload(int, float)
  @overload(int, int)
  def __init__(self, realPart: float, imagPart: float) -> None:
    """
    This constructor is overloaded with any combination of int and float.
    Since the overload decorator returns the function object without
    changes, as many overloads as needed can be added.
    """
    self.REAL = float(realPart)
    self.IMAG = float(imagPart)

  @overload(complex)  # Instantiating from a complex number instead
  def __init__(self, complexNumber: complex) -> None:
    """
    To instantiate from a complex number, a different function
    implementation is necessary.
    """
    self.REAL = complexNumber.real
    self.IMAG = complexNumber.imag

  @overload(THIS)  # THIS?
  def __init__(self, other: Self) -> None:
    """
    But what if you wanted to instantiate the class from another instance
    of the class? The moment this function object is created, the owning
    class itself does not actually exist yet. 'worktoy' provides the
    special token object 'THIS' to indicate the class yet to be created.
    """
    self.REAL = other.REAL
    self.IMAG = other.IMAG

  @overload()  # No arguments
  def __init__(self, **kwargs) -> None:
    """
    This constructor is called when no positional arguments are given. The
    positional arguments determine which constructor is called. The above
    constructors do not support keyword arguments, but this one does.
    """
    if 'real' in kwargs:
      self.REAL = kwargs['real']
    if 'imag' in kwargs:
      self.IMAG = kwargs['imag']

  #  Virtual accessor methods
  @ABS.GET
  def _getAbs(self):
    """
    The @ABS.GET decorator specifies that this method is the getter
    for the ABS property. When the ABS property is accessed through an
    instance of the class, this method is called.
    """
    return abs(self)  # Yes, we implement __abs__ further down

  @ABS.SET
  def _setAbs(self, value: float) -> None:
    """
    But how does one 'set' a virtual attribute? Well, however one
    wants! In this case, we scale the number to the new absolute value.
    """
    if not self:  # Yes, __bool__ is implemented further down
      raise ZeroDivisionError
    scale = value / abs(self)
    self.REAL *= scale
    self.IMAG *= scale

  @ARG.GET
  def _getArg(self) -> float:
    """
    The ARG property is the argument of the complex number. The @ARG.GET
    decorator specifies this method as getter.
    """
    return atan2(self.IMAG, self.REAL)  # With math.atan2

  @ARG.SET
  def _setArg(self, value: float) -> None:
    """
    The @ARG.SET decorator specifies this method as setter for the ARG
    property. The argument is set by rotating the complex number to the
    new angle.
    """
    if not self:
      raise ZeroDivisionError
    self.REAL = self.ABS * cos(value)
    self.IMAG = self.ABS * sin(value)

  #  Bonus dunder methods as promised

  def __abs__(self, ) -> float:
    """
    The __abs__ method is called when the built-in abs() function is
    called on an instance of the class. It returns the absolute value
    of the complex number.
    """
    return (self.REAL ** 2 + self.IMAG ** 2) ** 0.5

  def __bool__(self, ) -> bool:
    """
    The __bool__ method is called when the built-in bool() function is
    called on an instance of the class. It returns True if the complex
    number is not zero, and False otherwise.
    """
    return True if self.REAL ** 2 + self.IMAG ** 2 > 1e-12 else False

  def __complex__(self, ) -> complex:
    """
    The __complex__ method is called when the built-in complex() function
    is called on an instance of the class. It returns the complex number
    as a complex object.
    """
    return self.REAL + self.IMAG * 1j

  def __str__(self, ) -> str:
    """
    This returns a human-readable string representation of the complex
    number.
    """
    return """%.3f + %.3fJ""" % (self.REAL, self.IMAG)  # Yes, 'J'

  def __repr__(self, ) -> str:
    """
    The string returned by this method should ideally be a valid Python
    expression that recreates the 'self' object when passed to 'eval()'.
    """
    clsName = type(self).__name__  # Get the class name
    x, y = self.REAL, self.IMAG  # Get the real and imaginary parts
    return """%s(%s, %s)""" % (clsName, x, y)

  #  Further dunder methods are left as an exercise to the try-hard readers.


def main(*args) -> int:
  """In the following, we will test the class. """
  # Let's create a few complex numbers
  z1 = ComplexNumber(69, 420)  # two ints
  z2 = ComplexNumber(0.1337, 80085)  # a float and an int
  z3 = ComplexNumber(z1)  # another complex number
  z4 = ComplexNumber(69 + 420 * 1j)  # a complex number
  z5 = ComplexNumber(real=1337, imag=420)  # from keyword arguments
  z0 = ComplexNumber()  # no arguments
  Z = [z0, z1, z2, z3, z4, z5]  # list of complex numbers
  print("""Testing ComplexNumber""")
  for i, z in enumerate(Z):
    if i:
      print(' --- ')
    print('ComplexNumber: z%d = %s' % (i, repr(z)))
    try:
      if (z.ABS - abs(complex(z))) ** 2 > 1e-10:
        raise ValueError('ABS should be equal to abs(complex(z)')
      if (z.REAL - complex(z).real) ** 2 > 1e-10:
        raise ValueError('REAL should be equal to complex(z).real')
      if (z.IMAG - complex(z).imag) ** 2 > 1e-10:
        raise ValueError('IMAG should be equal to complex(z).imag')

    except ValueError as valueError:
      print('oh oh!', valueError)
      break  # Better stop and do some debugging!
    else:
      print('Nice, no error!')
    finally:
      print('That was: %s' % z)
  else:
    print('\nNice, no errors!')
    return 0
  print('Derp!')
  return 1


if __name__ == '__main__':
  sys.exit(main(sys.argv))



"""AbstractField provides an implementation of the descriptor protocol
that allow the owning class to explicitly define the accessor methods.  """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..mcls import FunctionType as Func
from ..parse import maybe
from ..text import typeMsg
from ..waitaminute import VariableNotNone, MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Callable


class Field:
  """AbstractField provides an implementation of the descriptor protocol
  that allow the owning class to explicitly define the accessor methods.  """

  __field_name__ = None
  __field_owner__ = None

  __setter_keys__ = None
  __getter_key__ = None
  __deleter_keys__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def _getSetterKeys(self, ) -> tuple[str, ...]:
    """Getter-function for names of methods called when __set__ is called."""
    return maybe(self.__setter_keys__, ())

  def _getGetterKey(self, ) -> str:
    """Getter-function for name of method called when __get__ is called."""
    return self.__getter_key__

  def _getDeleterKeys(self) -> tuple[str, ...]:
    """Getter-function for names of methods called when __delete__ is
    called."""
    return maybe(self.__deleter_keys__, ())

  def _addSetterKey(self, setterKey: str) -> None:
    """Add a setter key to the list of setter keys."""
    if not isinstance(setterKey, str):
      raise TypeError(typeMsg('setterKey', setterKey, str))
    existing = self._getSetterKeys()
    self.__setter_keys__ = (*[*existing, setterKey],)

  def _setGetterKey(self, getterKey: str) -> None:
    """Set the getter key."""
    if self.__getter_key__ is not None:
      raise VariableNotNone('__getter_key__', )
    if not isinstance(getterKey, str):
      raise TypeError(typeMsg('getterKey', getterKey, str))
    self.__getter_key__ = getterKey

  def _addDeleterKey(self, key: str) -> None:
    """Add a deleter key to the list of deleter keys."""
    if not isinstance(key, str):
      raise TypeError(typeMsg('key', key, str))
    existing = self._getDeleterKeys()
    self.__deleter_keys__ = (*[*existing, key],)

  def __get__(self, instance: object, owner: type) -> Any:
    """Getter-function for the field."""
    if instance is None:
      return self
    getterKey = self._getGetterKey()
    if getterKey is None:
      raise MissingVariable('__getter_key__', str)
    if not isinstance(getterKey, str):
      raise TypeError(typeMsg('getterKey', getterKey, str))
    getterFunction = getattr(owner, getterKey, None)
    if getterFunction is None:
      raise MissingVariable(getterKey, Callable)
    if not callable(getterFunction):
      raise TypeError(typeMsg(getterKey, getterFunction, Func))
    return getterFunction(instance, )

  def __set__(self, instance: object, value: Any) -> None:
    """Setter-function for the field."""
    owner = type(instance)
    setterKeys = self._getSetterKeys()
    if not setterKeys:
      raise MissingVariable('__setter_keys__', tuple)
    if not isinstance(setterKeys, tuple):
      raise TypeError(typeMsg('__setter_keys__', setterKeys, tuple))
    for setterKey in setterKeys:
      if not isinstance(setterKey, str):
        raise TypeError(typeMsg('setterKey', setterKey, str))
      setterFunction = getattr(owner, setterKey, None)
      if setterFunction is None:
        raise MissingVariable(setterKey, Func)
      if not callable(setterFunction):
        raise TypeError(typeMsg(setterKey, setterFunction, Func))
      setterFunction(instance, value)

  def __delete__(self, instance: object) -> None:
    """Deleter-function for the field."""
    owner = type(instance)
    deleterKeys = self._getDeleterKeys()
    if not deleterKeys:
      raise MissingVariable('__deleter_keys__', tuple)
    if not isinstance(deleterKeys, tuple):
      raise TypeError(typeMsg('__deleter_keys__', deleterKeys, tuple))
    for deleterKey in deleterKeys:
      if not isinstance(deleterKey, str):
        raise TypeError(typeMsg('deleterKey', deleterKey, str))
      deleterFunction = getattr(owner, deleterKey, None)
      if deleterFunction is None:
        raise MissingVariable(deleterKey, Func)
      if not callable(deleterFunction):
        raise TypeError(typeMsg(deleterKey, deleterFunction, Func))
      deleterFunction(instance)

  def GET(self, callMeMaybe: Func) -> Func:
    """Decorator specifying the getter method for the field."""
    self._setGetterKey(callMeMaybe.__name__)
    return callMeMaybe

  def SET(self, callMeMaybe: Func) -> Func:
    """Decorator specifying the setter method for the field."""
    self._addSetterKey(callMeMaybe.__name__)
    return callMeMaybe

  def DELETE(self, callMeMaybe: Func) -> Func:
    """Decorator specifying the deleter method for the field."""
    self._addDeleterKey(callMeMaybe.__name__)
    return callMeMaybe

  def __init__(self, *args) -> None:
    if args:
      if isinstance(args[0], str):
        self._setGetterKey(args[0])
